from dune.perftool.generation import (get_backend,
                                      temporary_variable,
                                      instruction,
                                      preamble)
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.options import (get_form_option,
                                   option_switch)
from dune.perftool.pdelab.geometry import (name_jacobian_determinant,
                                           world_dimension,
                                           local_dimension,
                                           apply_to_global_transformation,
                                           name_facet_jacobian_determinant,
                                           name_element_corners)
from dune.perftool.blockstructured.tools import sub_element_inames
import pymbolic.primitives as prim
from loopy.match import Writes


def compute_jacobian_2d(name):
    pos = get_backend("quad_pos", selector=option_switch("blockstructured"))()
    corners = name_element_corners()
    return instruction(expression=prim.Subscript(prim.Variable(corners), (0,0)),
                       assignee=prim.Subscript(prim.Variable(name), (0,0)),
                       within_inames=frozenset(sub_element_inames()+get_backend(interface="quad_inames")()),
                       depends_on=frozenset({Writes(corners), Writes(get_pymbolic_basename(pos))})
                       )



def define_jacobian_matrix(name):
    temporary_variable(name, shape_impl=('fm',), shape=(world_dimension(), world_dimension()))
    if world_dimension() == 2:
        compute_jacobian_2d(name)
    else:
        raise NotImplementedError()


def name_jacobian_matrix():
    name = "jacobian"
    define_jacobian_matrix(name)
    return name


def pymbolic_compute_determinant(matrix):
    pass


def define_jacobian_determinant(name):
    temporary_variable(name, shape=())
    jacobian = name_jacobian_matrix()
    determinant = pymbolic_compute_determinant(jacobian)
    return instruction(expression=prim.Subscript(prim.Variable(jacobian), (0,0)),
                       assignee=prim.Variable(name),
                       within_inames=frozenset(sub_element_inames()+get_backend(interface="quad_inames")()),
                       depends_on=frozenset({Writes(jacobian)})
                       )


def name_jacobian_determinant():
    name = "detjac"
    define_jacobian_determinant(name)
    return name


# scale determinant according to the order of the blockstructure
def pymbolic_jacobian_determinant():
    return prim.Quotient(prim.Variable(name_jacobian_determinant()),
                         prim.Power(get_form_option("number_of_blocks"), local_dimension()))


# scale Jacobian according to the order of the blockstructure
def pymbolic_jacobian_inverse(i, j, restriction):
    name_jit = get_backend(interface="name_jit", selector=option_switch("constant_transformation_matrix"))(restriction)
    return prim.Product((get_form_option("number_of_blocks"),
                         prim.Subscript(prim.Variable(name_jit), (j, i))))


# scale determinant according to the order of the blockstructure
def pymbolic_facet_jacobian_determinant():
    return prim.Quotient(prim.Variable(name_facet_jacobian_determinant()),
                         prim.Power(get_form_option("number_of_blocks"), local_dimension()))


# translate a point in the micro element into macro coordinates
def define_point_in_macro(name, point_in_micro):
    dim = local_dimension()
    temporary_variable(name, shape_impl=('fv',), shape=(dim,))

    # point_macro = (point_micro + index_micro) / number_of_blocks
    # where index_micro = tensor index of the micro element
    subelem_inames = sub_element_inames()
    for i in range(dim):
        if isinstance(point_in_micro, prim.Subscript):
            expr = prim.Subscript(point_in_micro.aggregate, point_in_micro.index + (i,))
        else:
            expr = prim.Subscript(point_in_micro, (i,))
        expr = prim.Sum((expr, prim.Variable(subelem_inames[i]),))
        expr = prim.Quotient(expr, get_form_option('number_of_blocks'))
        # TODO relax within inames
        instruction(assignee=prim.Subscript(prim.Variable(name), (i,)),
                    expression=expr,
                    within_inames=frozenset(subelem_inames),
                    )


# TODO add subelem inames if this function gets called
# TODO change input parameter to string
def name_point_in_macro(point_in_micro):
    assert isinstance(point_in_micro, prim.Expression)
    name = get_pymbolic_basename(point_in_micro) + "_macro"
    define_point_in_macro(name, point_in_micro)
    return name


def to_global(local):
    macro = name_point_in_macro(local)
    name = macro + "_global"
    apply_to_global_transformation(name, prim.Variable(macro))
    return prim.Variable(name)
